package at.bestsolution.rsd.rt.client.ejb.opentracing;

import java.util.ArrayDeque;
import java.util.Deque;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.TreeSet;

import org.eclipse.microprofile.config.ConfigProvider;
import org.osgi.service.component.annotations.Component;
import org.osgi.service.component.annotations.Reference;
import org.osgi.service.component.annotations.ReferenceCardinality;
import org.osgi.service.component.annotations.ReferencePolicy;
import org.osgi.service.component.annotations.ReferencePolicyOption;

import at.bestsolution.rsd.rt.client.ejb.EJBCallInterceptor;
import at.bestsolution.rsd.rt.client.ejb.EJBResultInterceptor;
import at.bestsolution.rsd.rt.opentracing.TraceProvider;
import io.opentracing.Scope;
import io.opentracing.Span;
import io.opentracing.SpanContext;
import io.opentracing.Tracer;
import io.opentracing.propagation.Format;
import io.opentracing.propagation.TextMap;
import io.opentracing.tag.Tags;
import io.opentracing.util.GlobalTracer;

@Component
public class OpenTracingClientEJBInterceptor implements EJBCallInterceptor, EJBResultInterceptor {
	
	private static final ThreadLocal<Deque<Scope>> ACTIVE_SCOPE = ThreadLocal.withInitial( () -> new ArrayDeque<>());
	
	private String getTraceProviderId() {
		return ConfigProvider.getConfig().getOptionalValue("rsd.opentracing.provider.id", String.class)
				.orElse("jaeger");
	}
	
	static class SortEntry implements Comparable<SortEntry> {
		private final int rank;
		private final TraceProvider tracer;
		
		public SortEntry(int rank, TraceProvider tracer) {
			this.rank = rank;
			this.tracer = tracer;
		}
		
		@Override
		public int compareTo(SortEntry o) {
			return Integer.compare(rank, o.rank) * -1;
		}
	}
	
	TreeSet<SortEntry> reg = new TreeSet<>();
	
	@Reference(cardinality = ReferenceCardinality.MULTIPLE, policy = ReferencePolicy.DYNAMIC, policyOption = ReferencePolicyOption.GREEDY)
	public void registerTraceProvider(TraceProvider tracer, Map<String, Object> properties) {
		Integer value = (Integer) properties.getOrDefault("service.ranking", 0);
		reg.add(new SortEntry(value, tracer));
	}
	
	public void unregisterTraceProvider(TraceProvider tracer) {
		reg.removeIf( e -> e.tracer == tracer);
	}
	
	public synchronized TraceProvider getActiveTracer() {
		if( this.reg.isEmpty() ) {
			return new TraceProvider() {
				
				@Override
				public Tracer getTracer() {
					return GlobalTracer.get();
				}
				
				@Override
				public String getId() {
					return "<unknown>";
				}
			};
		} else {
			String providerId = getTraceProviderId();
			
			if( providerId != null ) {
				Optional<TraceProvider> provider = this.reg
					.stream()
					.map(e -> e.tracer)
					.filter( p -> providerId.equals(p.getId()))
					.findFirst();
				
				if( provider.isPresent() ) {
					return provider.get();
				}
			}
			
			return this.reg.iterator().next().tracer;
		}
	}
	
	@Override
	public String name() {
		return "Opentracing-Interceptor";
	}
	
	// https://github.com/eclipse/microprofile-opentracing/blob/master/spec/src/main/asciidoc/microprofile-opentracing.asciidoc
	@Override
	public boolean preInvokation(EJBCallInterceptor.CallContext context) {
		try {
			TraceProvider provider = getActiveTracer();
			Tracer tracer = provider.getTracer();
			
			Tracer.SpanBuilder spanBuilder = tracer.buildSpan( context.getMethod().split("#")[1] );
			spanBuilder = spanBuilder.withTag(Tags.COMPONENT.getKey(), "ejb");
			spanBuilder = spanBuilder.withTag("ejb.bean", context.getMethod().split("#")[0]);
			spanBuilder = spanBuilder.withTag(Tags.SPAN_KIND.getKey(), Tags.SPAN_KIND_CLIENT);
			
			Span activeSpan = tracer.activeSpan();
			if( activeSpan != null ) {
				spanBuilder = spanBuilder.asChildOf(activeSpan.context());
			}
			
			Scope scope = spanBuilder.startActive(true);
			SpanContext ctx = scope.span().context();
			
			Map<String, String> map = new HashMap<>();
			
			tracer.inject(ctx, Format.Builtin.TEXT_MAP, new TextMap() {
				
				@Override
				public void put(String key, String value) {
					map.put(key, value);
				}
				
				@Override
				public Iterator<Entry<String, String>> iterator() {
					throw new IllegalStateException();
				}
			});
			
			context.setProperty("__rsd_trace-span-context", map);
			context.setProperty("__rsd_trace-provider-id", provider.getId());
			
			ACTIVE_SCOPE.get().push(scope);
		} catch( Throwable t ) {
			t.printStackTrace();
		}
		
		return true;
	}
	

	@Override
	public Object process(ResultContext context, Object result) {
		Scope scope = ACTIVE_SCOPE.get().poll();
		if( scope != null ) {
			scope.close();
		}
		
		return result;
	}
}
